///GPCA Header.

#define CMD_PAUSE_RESUME 0x02
#define PKT_SIZE 10

///GPCA Event (From Environment).
#define CMD_EVT_RequestBolus 0x40
#define CMD_EVT_StopInfusion 0x41
#define CMD_EVT_PowerButton 0x42
#define CMD_EVT_PauseInfusion 0x43
#define CMD_EVT_ConfirmPauseInfusion 0x44
#define CMD_EVT_ClearAlarm 0x45
#define CMD_EVT_Cancel 0x46
#define CMD_EVT_ChangeDoseRate 0x47
#define CMD_EVT_ConfigureInfusionProgram 0x48
#define CMD_EVT_ConfirmStopInfusion 0x49
#define CMD_EVT_ConfirmDoseRate 0x4A
#define CMD_EVT_NewInfusion 0x4B
#define CMD_EVT_ChangeVTBI 0x4C
#define CMD_EVT_ConfirmVTBI 0x4D
#define CMD_EVT_StartInfusion 0x4E
#define CMD_EVT_CheckAdminSet 0x4F
#define CMD_EVT_CheckAdminSet_1 0x50
#define CMD_EVT_CheckDrug 0x51
#define CMD_EVT_ConfirmConcentration 0x52
#define CMD_EVT_Prime 0x53

///GPCA Condition (From Environment).
#define CMD_V_T_Cond_6_6 0x70
#define CMD_V_T_Cond_6_3 0x71
#define CMD_V_T_Cond_6_4 0x72
#define CMD_V_T_Cond_6_5 0x73
#define CMD_V_MCond_6_6 0x74
#define CMD_V_TCond_6_3 0x75
#define CMD_V_Cond_6_3 0x76
#define CMD_V_TCond_6_5 0x77
#define CMD_V_TCond_6_4 0x78
#define CMD_V_Cond_7_4 0x79
#define CMD_V_Cond_7_3 0x7A
#define CMD_V_Cond_2 0x7B
#define CMD_V_Level_Two_Alarm 0x7C
#define CMD_V_Cond_4_1 0x7D
#define CMD_V_Cond_4_2 0x7E
#define CMD_V_Cond_4_3 0x7F
#define CMD_V_Cond_4_4 0x80
#define CMD_V_Cond_4_5 0x81
#define CMD_V_Cond_4_6 0x82
//#define CMD_V_Cond_5 0x83
#define CMD_V_Cond_5_1 0x83     //added.
#define CMD_V_MCDR_In_Progress 0x84
#define CMD_V_Cond_3_1 0x85
#define CMD_V_Cond_3_2 0x86
#define CMD_V_Cond_3_3 0x87
#define CMD_V_Cond_3_4 0x88
#define CMD_V_Cond_3_5 0x89
#define CMD_V_Cond_3_6 0x8A
#define CMD_V_Cond_3_7 0x8B
#define CMD_V_Cond_3_8 0x8C
#define CMD_V_Cond_3_9 0x8D
#define CMD_V_Cond_3_11 0x8E
#define CMD_V_Cond_3_10 0x8F
#define CMD_V_Power_ON 0x90
//#define CMD_V_MAX_POST_WAIT 0x91
#define CMD_V_MCond_1_3 0x92
#define CMD_V_MCond_1_4 0x93  //added.
#define CMD_V_Cond_5_2 0x94   //added.
#define CMD_V_Cond_6_2 0x95   //added
#define CMD_V_Cond_7_1 0x96   //added.
#define CMD_V_Cond_7_2 0x97   //added.

//Network constant
#define CHAN_ACK 1
#define CHAN_NO_ACK 0

///Packet index (Downstream from client to GPCA).
///Motor control packet constants.
#define INDEX_CTL_SOP 0
#define INDEX_CTL_CMD 1
#define INDEX_CTL_MOTOR_DIR 2
#define INDEX_CTL_MOTOR_SPD 3
#define INDEX_CTL_MOTOR_TIMEOUT1 4
#define INDEX_CTL_MOTOR_TIMEOUT2 5
#define INDEX_CTL_EOP 9
///Alarm control packet constants.
#define INDEX_CTL_ALARM_LED 2
#define INDEX_CTL_ALARM_BUZZ 3
///GPCA control packet constants.
#define INDEX_CTL_GPCA_CONDITION 2
#define INDEX_CTL_GPCA_EVENT 3

#define CTL_SOP 0xFA
#define CTL_EOP 0xFB
#define CTL_CMD_MOTOR 0x01
#define CTL_CMD_ALARM 0x02
#define CTL_CMD_GPCA 0x03
/**
	Parsing state (Downstream from GPCA to client).
*/
#define PARSE_CTL_INIT 0
#define PARSE_CTL_SOP_RCVD 1
#define PARSE_CTL_CMD_MOTOR_RCVD 2
#define PARSE_CTL_CMD_ALARM_RCVD 3
#define PARSE_CTL_CMD_GPCA_RCVD 4
#define PARSE_CTL_MOTOR_DIR_RCVD 5
#define PARSE_CTL_MOTOR_SPD_RCVD 6
#define PARSE_CTL_MOTOR_TIMEOUT1_RCVD 7
#define PARSE_CTL_MOTOR_TIMEOUT2_RCVD 8
#define PARSE_CTL_ALARM_LED_RCVD 9
#define PARSE_CTL_ALARM_BUZZ_RCVD 10
#define PARSE_CTL_ALARM_GPCA_CONDITION_RCVD 11
#define PARSE_CTL_ALARM_GPCA_EVENT_RCVD 12
#define PARSE_CTL_EOP_RCVD 13


///Packet index (Upstream from GPCA to client).
#define INDEX_SOP 0
#define INDEX_MOTOR_DIR 1
#define INDEX_SPD_LEVEL 2
#define INDEX_DURATION 3
#define INDEX_GPCA_STATE 4  
#define INDEX_MOTOR_STATE 5
#define INDEX_CHANNEL_ACK 6
#define INDEX_REMAIN_VOL_PERCENT 7
#define INDEX_EOP 9
#define SOP     0xFC
#define EOP     0xFD

///Motor-Control related constants.
#define CLOCKWISE 0
#define COUNTER_CLOCKWISE 1
#define MAX_MOTOR_SPEED 5
#define MOTOR_STOP      0
#define MOTOR_RUN       1

/**
	Parsing related (Upstream from GPCA to client).
*/
#define PARSE_INIT 0
#define PARSE_SOP_RCVD 1
#define PARSE_MOTOR_DIR_RCVD 2
#define PARSE_SPD_LEVEL_RCVD 3
#define PARSE_DURATION_RCVD 4
#define PARSE_GPCA_STATE_RCVD 5
#define PARSE_MOTOR_STATE_RCVD 6
#define PARSE_RESERVE6_RCVD 7
#define PARSE_RESERVE7_RCVD 8
#define PARSE_RESERVE8_RCVD 9
#define PARSE_EOF_RCVD 10

/**
	Controller commands.
*/
#define CONT_CMD_TOGGLE_MOTOR_DIRECTION 0x01
#define CONT_CMD_TOGGLE_MOTOR_OPERATION 0x02	//Stop or start.
#define CONT_CMD_SET_MOTOR_SPD_LEVEL1 0x10
#define CONT_CMD_SET_MOTOR_SPD_LEVEL2 0x11
#define CONT_CMD_SET_MOTOR_SPD_LEVEL3 0x12
#define CONT_CMD_SET_MOTOR_SPD_LEVEL4 0x13

#define CONT_CMD_SET_MOTOR_DURATION_LEVEL1 0x20
#define CONT_CMD_SET_MOTOR_DURATION_LEVEL2 0x21
#define CONT_CMD_SET_MOTOR_DURATION_LEVEL3 0x22
#define CONT_CMD_SET_MOTOR_DURATION_LEVEL4 0x23
#define CONT_CMD_SET_MOTOR_DURATION_LEVEL5 0x24
#define CONT_CMD_SET_MOTOR_DURATION_LEVEL6 0x25

#define CONT_CMD_ALARM_BUZZ_ON 0x30
#define CONT_CMD_ALARM_BUZZ_OFF 0x31
#define CONT_CMD_ALARM_LED_ON 0x32
#define CONT_CMD_ALARM_LED_OFF 0x33


