#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>

#include "times_header.h"
#include "network_header.h"
#include "gpca_header.h"

/** @defgroup TCP_Network_Variables TCP Connection-related variables.
* @{
*/
/// Server port number.
int servPort = 0; 
///Server Socket Descriptor.
int sockfd = 0;   
///Client Socket Descriptor.
int clnt_socket = 0;	
///Server and Client address structure.
struct sockaddr_in serv_addr, clnt_addr;
///socket writing buffer.
char sock_write_buf[SOCK_WRITE_BUF_SIZE];		
///socket read buffer.
char sock_read_buf[SOCK_READ_BUF_SIZE];		

/** @} */  //end of group

/** @defgroup Serial_Variables Serial Connection-related variables.
* @{
*/
///Serial File Descriptor.
int serial_fd;	
///Serial read buffer.
unsigned char serial_read_buf[SERIAL_READ_BUF_SIZE];		
char buf[1024];
///Contain a serial device name (e.g. /dev/ttyUSB0).
char devicename[80];	

/** @} */ //end of group

/** @defgroup Thread_functions Thread functions.
* @{
*/
///Manages streams from socket clients to GPCA-StateMachine.
void *downstream_thread (void *ptr); 
///Manages streams from GPCA-Controller to GPCA-StateMachine.
void *upstream_thread (void *ptr);  
///GPCA state machine thread :Mainly execute TIMES codes.
void *gpca_state_machine_thread (void *ptr);  
///GPCA clock manager
void *gpca_clock_manager (void *ptr);  
///Observer adapter.
void *observer_adapter (void *ptr);	

/** @} */ //end of group

/** @defgroup TIMES_Interface Automatically generated TIMES interface.
* @{
*/
///TIMES interface.
static void clear_and_set ( unsigned char sync ); 
///TIMES interface.
void setClock(time_t clockID, long value);  
///TIMES interface.
time_t rdClock(time_t clockID);	
///TIMES interface.
void assign(int trn);		
///TIMES interface.
bool eval_guard(int trn);
///TIMES interface.
void check_trans(void);		

/**
 * Clock variables
 * Ordered: global first, then local clocks for each process.
 */
//time_t clock_GPCA_Instance_clk_x1;
time_t GPCA_Instance_clk_x1;

/**
 * Integer variables
 */

///An event channel for Bolus request from patients.
int E_RequestBolus=0;
///An event channel for pushing [stop button] by patients.
int E_StopInfusion=0;
///Pressing of the power button.
int E_PowerButton=0;
///An event channel for pushing [pause button] by patients.
int E_PauseInfusion=0;
///An event channel for pushing [confirm pause] by patients.
int E_ConfirmPauseInfusion=0;
///An event channel for clearing the current alarm by patients or caregivers.
int E_ClearAlarm=0;
///An event channel for cancelling the previous actions done by patients or caregivers.
int E_Cancel=0;
///An event channel for pushing [Change Dose Rate] button by patients or caregivers.
int E_ChangeDoseRate=0;
///The action of programming infusion parameters.
int E_ConfigureInfusionProgram=0;
///An event channel for pushing [Confirm Stop Infusion] by patients or caregivers.
int E_ConfirmStopInfusion=0;
///An event channel for pushing [Confirm Dose Rate] by patients or caregivers.
int E_ConfirmDoseRate=0;
///An event channel for initiating [infusion session submachine] after pump is being stopped.
int E_NewInfusion=0;
///Request to change the current VTBI.
int E_ChangeVTBI=0;
///The action of accepting the currently programmed VTBI.
int E_ConfirmVTBI=0;
///The action of accepting the current programmed parameters and starting the infusion.
int E_StartInfusion=0;
///Request for an administration set check.
int E_CheckAdminSet=0;
int E_CheckAdminSet_1=0;
///Request to check the loaded drug against the prescription.
int E_CheckDrug=0;
///The action of accepting the concentration of the loaded drug.
int E_ConfirmConcentration=0;
///Request to prime the pump.
int E_Prime=0;
///An event channel for indicating [Warning Low Reservoir].
int T_Cond_6_6=0;
///An event channel for indicating [Alarm Empty Reservoir].
int T_Cond_6_3=0;
///An event channel for indicating a new valid syringe has been inserted.
int T_Cond_6_4=0;
///An event channel for indicating an invalid syringe has been inserted.
int T_Cond_6_5=0;
///A condition variable for indicating [Wrn_LowReservior].
int MCond_6_6=0;
int TCond_6_3=0;
///A condition variable for indicating [Alrm_EmptyReservior].
int Cond_6_3=0;
///A condition variable for indicating [Alrm_WrongDrug].
int TCond_6_5=0;
///A condition variable for indicating that the reloaded drug matches the drug expected by the current infusion session.
int TCond_6_4=0;
///A condition variable for indicating that the new infusion rate that the users input during the current infusion satisfies both soft and hard limits suggested by the drug library.
int Cond_7_4=0;
///A condition variable for indicating that the new infusion rate that the user inputs during the current infusion violates either soft or hard limits suggested by the drug library.
int Cond_7_3=0;
///A condition variable for indicating that the pump is functioning properly and all necessary parts are ready for the infusion to start.
int Cond_2=0;
///An event channel for indicating [Level Two Alarm].
int Level_Two_Alarm=0;
///The value of VTBI provided by the user violates hard limits specified by the drug library. 
int Cond_4_1=0;
///The value of VTBI provided by the user violates soft limits specified by the drug library.
int Cond_4_2=0;
///The VTBI provided by the user complies with that prescribed by the drug library.
int Cond_4_3=0;
///The value of infusion dose rate provided by the user violates hard limits specified by the drug library.
int Cond_4_4=0;
///The value of infusion dose rate provided by the user violates soft limits specified by the drug library.
int Cond_4_5=0;
///The infusion dose rate provided by the user complies with that prescribed by the drug library.
int Cond_4_6=0;
int Cond_5_1=0;
int MCDR_In_Progress=0;
///Administration Check failed.
int Cond_3_1=0;
///Administration Check passed.
int Cond_3_2=0;
///Prime Check failed.
int Cond_3_3=0;
///Prime Check passed.
int Cond_3_4=0;
///The drug label does not reflect correctly and precisely the drug in the reservoir, or the drug is not prescribed by the drug library.
int Cond_3_5=0;
///The drug label reflects correctly and precisely the drug in the reservoir.
int Cond_3_6=0;
///The drug dose unit read by the barcode reader does not match that prescribed by the drug library.
int Cond_3_7=0;
///The drug dose unit read by the barcode reader matches that prescribed by the drug library.
int Cond_3_8=0;
///The drug concentration read by the barcode reader violates soft limits suggested by the drug library.
int Cond_3_9=0;
///The drug concentration read by the barcode reader complies with that prescribed by the drug library.
int Cond_3_11=0;
///The drug concentration read by the barcode reader violates hard limits suggested by the drug library.
int Cond_3_10=0;
int Power_ON=0;
int MCond_1_3=0;
int MCond_1_4=0;
int Cond_5_2=0;
int Cond_6_2=0;
int Cond_7_1=0;
int Cond_7_2=0;

/**
 * Transition table.
 */
trans_t trans[NB_TRANS] = {
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_ConfirmPause,-1},
  {false,GPCA_Instance_ConfirmPause,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_ConfirmPause,GPCA_Instance_InfusionPaused,-1},
  {false,GPCA_Instance_InfusionPaused,GPCA_Instance_Alrm_TooLongInfusionPause,-1},
  {false,GPCA_Instance_Alrm_TooLongInfusionPause,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Alrm_TooLongInfusionPause,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_ConfirmStop,-1},
  {false,GPCA_Instance_ConfirmStop,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_ConfirmStop,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_Alrm_LevelTwoHardwareFailure,-1},
  {false,GPCA_Instance_Alrm_LevelTwoHardwareFailure,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_Alrm_LevelTwoHardwareFailure,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_Wrn_LowReservior,-1},
  {false,GPCA_Instance_Wrn_LowReservior,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_Alrm_EmptyReservior,-1},
  {false,GPCA_Instance_Alrm_EmptyReservior,GPCA_Instance_CheckDrug,-1},
  {false,GPCA_Instance_CheckDrug,GPCA_Instance_Alrm_WrongDrug,-1},
  {false,GPCA_Instance_Alrm_WrongDrug,GPCA_Instance_CheckDrug,-1},
  {false,GPCA_Instance_CheckDrug,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_CheckDrug,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_ChangeRate,-1},
  {false,GPCA_Instance_ChangeRate,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_ChangeRate,GPCA_Instance_CheckNewRate,-1},
  {false,GPCA_Instance_CheckNewRate,GPCA_Instance_ChangeRate,-1},
  {false,GPCA_Instance_CheckNewRate,GPCA_Instance_Alrm_UnsafeNewRate,-1},
  {false,GPCA_Instance_Alrm_UnsafeNewRate,GPCA_Instance_Alrm_LevelTwoHardwareFailure,-1},
  {false,GPCA_Instance_DisplayVTBI,GPCA_Instance_ChangeVTBI,-1},
  {false,GPCA_Instance_ChangeVTBI,GPCA_Instance_DisplayVTBI,-1},
  {false,GPCA_Instance_ChangeVTBI,GPCA_Instance_Alrm_LongWait_ChangeVTBI,-1},
  {false,GPCA_Instance_Alrm_LongWait_ChangeVTBI,GPCA_Instance_ChangeVTBI,-1},
  {false,GPCA_Instance_CheckVTBI,GPCA_Instance_Alrm_VTBIHardLimitsViolation,-1},
  {false,GPCA_Instance_Alrm_VTBIHardLimitsViolation,GPCA_Instance_ChangeVTBI,-1},
  {false,GPCA_Instance_ChangeVTBI,GPCA_Instance_CheckVTBI,-1},
  {false,GPCA_Instance_CheckVTBI,GPCA_Instance_Wrn_VTBISoftLimitsViolation,-1},
  {false,GPCA_Instance_CheckVTBI,GPCA_Instance_DisplayDoseRate,-1},
  {false,GPCA_Instance_Wrn_VTBISoftLimitsViolation,GPCA_Instance_DisplayDoseRate,-1},
  {false,GPCA_Instance_DisplayVTBI,GPCA_Instance_DisplayDoseRate,-1},
  {false,GPCA_Instance_Wrn_VTBISoftLimitsViolation,GPCA_Instance_ChangeVTBI,-1},
  {false,GPCA_Instance_DisplayDoseRate,GPCA_Instance_ChangeDoseRate,-1},
  {false,GPCA_Instance_ChangeDoseRate,GPCA_Instance_DisplayDoseRate,-1},
  {false,GPCA_Instance_ChangeDoseRate,GPCA_Instance_Alrm_LongWait_ChangeDoseRate,-1},
  {false,GPCA_Instance_Alrm_LongWait_ChangeDoseRate,GPCA_Instance_ChangeDoseRate,-1},
  {false,GPCA_Instance_ChangeDoseRate,GPCA_Instance_CheckDoseRate,-1},
  {false,GPCA_Instance_CheckDoseRate,GPCA_Instance_Alarming_DoseRate_OutOfBound,-1},
  {false,GPCA_Instance_Alarming_DoseRate_OutOfBound,GPCA_Instance_ChangeDoseRate,-1},
  {false,GPCA_Instance_CheckDoseRate,GPCA_Instance_Wrn_DoseRateSoftLimitsViolation,-1},
  {false,GPCA_Instance_Wrn_DoseRateSoftLimitsViolation,GPCA_Instance_ChangeDoseRate,-1},
  {false,GPCA_Instance_CheckDoseRate,GPCA_Instance_DisplaySetting,-1},
  {false,GPCA_Instance_DisplayDoseRate,GPCA_Instance_DisplaySetting,-1},
  {false,GPCA_Instance_Wrn_DoseRateSoftLimitsViolation,GPCA_Instance_DisplaySetting,-1},
  {false,GPCA_Instance_DisplaySetting,GPCA_Instance_Alrm_LongWait_StartInfusion,-1},
  {false,GPCA_Instance_Alrm_LongWait_StartInfusion,GPCA_Instance_DisplaySetting,-1},
  {false,GPCA_Instance_DisplaySetting,GPCA_Instance_ReadyToStart,-1},
  {false,GPCA_Instance_ReadyToStart,GPCA_Instance_DisplaySetting,-1},
  {false,GPCA_Instance_CDR_Init,GPCA_Instance_CheckAdministrationSet,-1},
  {false,GPCA_Instance_CheckAdministrationSet,GPCA_Instance_Alrm_IncorrectAdministrationSetup,-1},
  {false,GPCA_Instance_Alrm_IncorrectAdministrationSetup,GPCA_Instance_CheckAdministrationSet,-1},
  {false,GPCA_Instance_CheckAdministrationSet,GPCA_Instance_CheckPrime,-1},
  {false,GPCA_Instance_CheckPrime,GPCA_Instance_Alrm_PrimeFailure,-1},
  {false,GPCA_Instance_Alrm_PrimeFailure,GPCA_Instance_CheckPrime,-1},
  {false,GPCA_Instance_CheckPrime,GPCA_Instance_DisplayDrugInfo,-1},
  {false,GPCA_Instance_DisplayDrugInfo,GPCA_Instance_CheckDrugType,-1},
  {false,GPCA_Instance_CheckDrugType,GPCA_Instance_Alrm_UnknownDrug,-1},
  {false,GPCA_Instance_Alrm_UnknownDrug,GPCA_Instance_CheckDrugType,-1},
  {false,GPCA_Instance_CheckDrugType,GPCA_Instance_CheckDoseUnit,-1},
  {false,GPCA_Instance_CheckDoseUnit,GPCA_Instance_Alrm_DoseUnitMismatch,-1},
  {false,GPCA_Instance_Alrm_DoseUnitMismatch,GPCA_Instance_CheckDrugType,-1},
  {false,GPCA_Instance_CheckDoseUnit,GPCA_Instance_CheckConcentration,-1},
  {false,GPCA_Instance_CheckConcentration,GPCA_Instance_Wrn_ConcentrationSoftLimitsViolation,-1},
  {false,GPCA_Instance_Wrn_ConcentrationSoftLimitsViolation,GPCA_Instance_CheckDrugType,-1},
  {false,GPCA_Instance_CheckConcentration,GPCA_Instance_DisplayPatientProfile,-1},
  {false,GPCA_Instance_Wrn_ConcentrationSoftLimitsViolation,GPCA_Instance_DisplayPatientProfile,-1},
  {false,GPCA_Instance_DisplayPatientProfile,GPCA_Instance_DisplayDrugInfo,-1},
  {false,GPCA_Instance_Alrm_ConcentrationHardLimitsViolation,GPCA_Instance_CheckDrugType,-1},
  {true,GPCA_Instance_POST_Init,GPCA_Instance_Post_In_Progress,-1},
  {false,GPCA_Instance_Post_In_Progress,GPCA_Instance_CDR_Init,-1},
  {false,GPCA_Instance_Post_In_Progress,GPCA_Instance_Alrm_POSTFailure,-1},
  {false,GPCA_Instance_Alrm_POSTFailure,GPCA_Instance_POST_Init,-1},
  {false,GPCA_Instance_Post_In_Progress,GPCA_Instance_Alrm_POSTFailure,-1},
  {false,GPCA_Instance_CheckConcentration,GPCA_Instance_Alrm_ConcentrationHardLimitsViolation,-1},
  {false,GPCA_Instance_DisplayPatientProfile,GPCA_Instance_DisplayVTBI,-1},
  {false,GPCA_Instance_ReadyToStart,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_InfusionStopped,GPCA_Instance_CDR_Init,-1},
  {false,GPCA_Instance_InfusionPaused,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_ConfirmPause,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_InfusionPaused,GPCA_Instance_PausedStopConfirm,-1},
  {false,GPCA_Instance_PausedStopConfirm,GPCA_Instance_InfusionPaused,-1},
  {false,GPCA_Instance_PausedStopConfirm,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Alrm_TooLongInfusionPause,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_ConfirmPause,GPCA_Instance_Alrm_LevelTwoHardwareFailure,-1},
  {false,GPCA_Instance_Alrm_LevelTwoHardwareFailure,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_ChangeRate,GPCA_Instance_Alrm_LevelTwoHardwareFailure,-1},
  {false,GPCA_Instance_ConfirmStop,GPCA_Instance_Alrm_LevelTwoHardwareFailure,-1},
  {false,GPCA_Instance_Alrm_UnsafeNewRate,GPCA_Instance_ChangeRate,-1},
  {false,GPCA_Instance_Alrm_UnsafeNewRate,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_ConfirmStop,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Alrm_UnsafeNewRate,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_InfusionStopped,GPCA_Instance_Alrm_LevelTwoHardwareFailure,-1},
  {false,GPCA_Instance_ChangeRate,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_ChangeRate,GPCA_Instance_Alrm_LongWait_ChangeDoseRate_ISSM,-1},
  {false,GPCA_Instance_Alrm_LongWait_ChangeDoseRate_ISSM,GPCA_Instance_ChangeRate,-1},
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Alrm_WrongDrug,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Alrm_EmptyReservior,GPCA_Instance_InfusionStopped,-1},
  {false,GPCA_Instance_Infusion_NormalOperation,GPCA_Instance_BolusRequest,-1},
  {false,GPCA_Instance_BolusRequest,GPCA_Instance_Infusion_NormalOperation,-1},
  {false,GPCA_Instance_BolusRequest,GPCA_Instance_Infusion_NormalOperation,-1}
};

/**
 * Location list
 */
loc_t loc[NB_TRANS+NB_LOC] = {
  0,6,9,12,14,20,101,104,tid_NOP/*Infusion_NormalOperation*/,
  1,2,84,89,tid_NOP/*ConfirmPause*/,
  3,83,85,tid_NOP/*InfusionPaused*/,
  4,5,88,tid_NOP/*Alrm_TooLongInfusionPause*/,
  82,97,tid_NOP/*InfusionStopped*/,
  7,8,92,95,tid_NOP/*ConfirmStop*/,
  10,11,90,tid_NOP/*Alrm_LevelTwoHardwareFailure*/,
  13,tid_NOP/*Wrn_LowReservior*/,
  15,103,tid_NOP/*Alrm_EmptyReservior*/,
  16,18,19,tid_NOP/*CheckDrug*/,
  17,102,tid_NOP/*Alrm_WrongDrug*/,
  21,22,91,98,99,tid_NOP/*ChangeRate*/,
  23,24,tid_NOP/*CheckNewRate*/,
  25,93,94,96,tid_NOP/*Alrm_UnsafeNewRate*/,
  26,36,tid_NOP/*DisplayVTBI*/,
  27,28,32,tid_NOP/*ChangeVTBI*/,
  29,tid_NOP/*Alrm_LongWait_ChangeVTBI*/,
  30,33,34,tid_NOP/*CheckVTBI*/,
  31,tid_NOP/*Alrm_VTBIHardLimitsViolation*/,
  35,37,tid_NOP/*Wrn_VTBISoftLimitsViolation*/,
  38,48,tid_NOP/*DisplayDoseRate*/,
  39,40,42,tid_NOP/*ChangeDoseRate*/,
  41,tid_NOP/*Alrm_LongWait_ChangeDoseRate*/,
  43,45,47,tid_NOP/*CheckDoseRate*/,
  44,tid_NOP/*Alarming_DoseRate_OutOfBound*/,
  46,49,tid_NOP/*Wrn_DoseRateSoftLimitsViolation*/,
  50,52,tid_NOP/*DisplaySetting*/,
  51,tid_NOP/*Alrm_LongWait_StartInfusion*/,
  53,81,tid_NOP/*ReadyToStart*/,
  72,80,tid_NOP/*DisplayPatientProfile*/,
  54,tid_NOP/*CDR_Init*/,
  55,57,tid_NOP/*CheckAdministrationSet*/,
  56,tid_NOP/*Alrm_IncorrectAdministrationSetup*/,
  58,60,tid_NOP/*CheckPrime*/,
  59,tid_NOP/*Alrm_PrimeFailure*/,
  61,tid_NOP/*DisplayDrugInfo*/,
  62,64,tid_NOP/*CheckDrugType*/,
  63,tid_NOP/*Alrm_UnknownDrug*/,
  65,67,tid_NOP/*CheckDoseUnit*/,
  66,tid_NOP/*Alrm_DoseUnitMismatch*/,
  68,70,79,tid_NOP/*CheckConcentration*/,
  69,71,tid_NOP/*Wrn_ConcentrationSoftLimitsViolation*/,
  73,tid_NOP/*Alrm_ConcentrationHardLimitsViolation*/,
  74,tid_NOP/*POST_Init*/,
  75,76,78,tid_NOP/*Post_In_Progress*/,
  77,tid_NOP/*Alrm_POSTFailure*/,
  86,87,tid_NOP/*PausedStopConfirm*/,
  100,tid_NOP/*Alrm_LongWait_ChangeDoseRate_ISSM*/,
  105,106,tid_NOP/*BolusRequest*/
};



/** @} */ // end of group

/** @defgroup Manual_GPCA_Interface Manually generated GPCA interface (Interact with TIMES code).
* @{
*/
///A global variable to keep track of the current GPCA state (Glue code).
int current_GPCA_state = GPCA_Instance_POST_Init;
int prev_GPCA_state = GPCA_Instance_POST_Init;
///An action handler on state change (Glue code).
int action_handler_on_state_change(int prev_state, int current_state);
///Return the current GPCA state (Glue code).
int getCurrentState(void); 
///Activate a channel based on the environmental stimulus.
int activate_gpca_channel(unsigned char cmd);
///Toggle a condition variable received from the environment.	
//int toggle_gpca_condition(unsigned char cmd);

/** @} */ // end of group

/**
* @defgroup GPCA_Protocol_Variables	GPCA Protocol related variables.
* @{
*/
///Pump motor speed : 1 (20ms), 2 (40ms), 3 (60ms) 5 (100ms).
uint8_t MOTOR_SPEED_LEVEL = 1;  
///Motor direction.
uint8_t MOTOR_DIRECTION = CLOCKWISE; 
///Motor state : Stop or Active.
uint8_t MOTOR_STATE = MOTOR_STOP;	
///Infusion duration : 1 (30sec), 2 (60sec), 3 (90sec), 4 (120 sec), 5 (150 sec), 6 (180 sec).
uint8_t INFUSION_DURATION = 1; 
///Remaining volume percent. 
uint8_t REMAIN_VOL_PERCENT;
///Buffer to contain gpca packets.
uint8_t pkt_gpca[PKT_SIZE];	
///Function to send gpca packets to client.
int send_gpca_pkt_to_client(void);  
///Fuction to send channel acks to client.
int send_channel_ack_to_client(void);
///Parsing state (upstream).
int parsing_state = PARSE_INIT;	
///Parsing state (downstream from client to GPCA).
int ctl_parsing_state = PARSE_CTL_INIT;
/** @} */ // end of group

///A function to send one-byte command to the infusion controller.
int write_byte_to_controller(int fd, char data);

/** \brief Main function.
*
*	\param Parm_Count the number of parameters passed to main function.
*	\param Parms[] 1) device name 2) Server port number.
*	\exception fail to open serial devices.
*	\return 0 
*/	
int main(int Parm_Count, char* Parms[]){
	int i;
	pthread_t thread_downstream, thread_upstream, thread_gpcaStateMachine, thread_clockManager, thread_observer_adapter;
	int iret1, iret2, iret3, iret4, iret5;

	///Check stdin parameters.
	if(Parm_Count != 3){
		printf("wrong parameters : [devicename] [server port]\n");
		return 0;
	}
	
	strcpy(buf, Parms[1]);
	i = sscanf (buf, "%s", devicename);
	if (i != 1)
		printf("Error : sscanf\n");

	///Open serial device connection.
	serial_fd = open(devicename, O_RDWR | O_NOCTTY);
	if (serial_fd < 0) {
		printf("Error : open\n");
		//exit (-1);
	}else{
		printf("Serial open successful : %s \n", devicename);
	}
	
	///Fill the address for server socket.
	servPort = atoi(Parms[2]);

	///Create threads for upstream and downstream independently. 
	iret1 = pthread_create(&thread_downstream, NULL, downstream_thread, NULL);
	iret2 = pthread_create(&thread_upstream, NULL, upstream_thread, NULL);
	iret3 = pthread_create(&thread_gpcaStateMachine, NULL, gpca_state_machine_thread, NULL);
	iret4 = pthread_create(&thread_clockManager, NULL, gpca_clock_manager, NULL);
	iret5 = pthread_create(&thread_observer_adapter, NULL, observer_adapter, NULL);

	pthread_join(thread_downstream, NULL);
	pthread_join(thread_upstream, NULL);
	pthread_join(thread_gpcaStateMachine, NULL);
	pthread_join(thread_clockManager, NULL);
	pthread_join(thread_observer_adapter, NULL);

	exit(0);
}

/** /brief thread that manages a downstream from client to GPCA-StateMachine.
*	\param ptr not used.
*	\return none.
*/
void *downstream_thread (void *ptr){

	int clnt_len;
	uint8_t read_error = 0;	///check socket read error.
	int read_res, write_res, channel_res, cond_res;

	printf("downstream_thread created\n");
	bzero((char*) &serv_addr, sizeof(serv_addr));
	bzero((char*) &clnt_addr, sizeof(clnt_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = INADDR_ANY;
	serv_addr.sin_port = htons(servPort);

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0){
		printf("Error opening socket\n\r");
	}
		
	if(bind(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0){
		printf("Error on binding\n\r");
	}

	while(1){
		listen(sockfd, 5);
		printf("Waiting for client connection on port [%d]...\n", servPort);
		clnt_len = sizeof(clnt_addr);
		clnt_socket = accept(sockfd, (struct sockaddr *) &clnt_addr, &clnt_len);
		if(clnt_socket < 0){
			printf("Error on accept\n\r");
			read_error = 1;
		}else{
			
			read_error = 0;
		}

		while(read_error == 0){
			///Initialize variable before starting.
			bzero((char*) sock_read_buf, SOCK_READ_BUF_SIZE);
			printf("Waiting for data from Client...\n\r");
			read_res = read(clnt_socket, sock_read_buf, 1);
			if(read_res <= 0){
				printf("Socket Error : READ \n\r");
				read_error = 1;
			}
			else{
				write_res = write(serial_fd, &sock_read_buf[0], 1);
				if(write_res <= 0){
					printf("Error in writing serial...\n\r");	
				}else{
					printf("[%c] was written\n\r", sock_read_buf[0]);
				}

				///Glue Code : Interface between Environment and GPCA (Channel communication).
				channel_res = activate_gpca_channel((unsigned char)sock_read_buf[0]);
				if(channel_res == 0){
					printf("Unknown channel or not in the valid state : %d\n", (unsigned char)sock_read_buf[0]);
				}
			}
		}//End while(for user input)
		printf("Client connection is closing : [%d]\n", clnt_socket);
		close(clnt_socket);	//Close client socket
	}//End while (for listen)
	
	close(serial_fd);	//Close serial port
	pthread_exit(NULL);	//Terminate thread

}

/** \brief A thread that manages upstreams from GPCA-Controller to GPCA-StateMachine
*	\param ptr not used.
*	\return none.
*/
void *upstream_thread(void *ptr){
	uint8_t serial_read_err = 0;	//check serial read error
	int read_res, write_res;
	uint8_t tempMotorDir, tempSpdLevel, tempDuration, tempMotorState, tempRemainVol;
	
	printf("upstream_thread created\n");
	while(1){
		serial_fd = open(devicename, O_RDWR | O_NOCTTY);
		if (serial_fd < 0) {
			printf("Error : open\n");
			serial_read_err = 1;
			sleep(1);
		}
		else{
			serial_read_err = 0;
		}	
	
		parsing_state = PARSE_INIT;		//initialize parsing state	
		while(serial_read_err == 0){

			bzero((char*)serial_read_buf, SERIAL_READ_BUF_SIZE);

			///Read data from the serial connection.
			read_res = read(serial_fd, serial_read_buf, 1);
			if(read_res <=0){
				printf("Serial Error :READ \n\r");
				serial_read_err = 1;
			}else{
				//Start Parsing for GPCA-packet
                		if((parsing_state == PARSE_INIT) && (serial_read_buf[0] == SOP)){
                    			parsing_state = PARSE_SOP_RCVD;
               			}
               			//Process Motor direction
               			else if((parsing_state == PARSE_SOP_RCVD)){
                    			parsing_state = PARSE_MOTOR_DIR_RCVD;
                    			tempMotorDir = serial_read_buf[0];
              			}                
             			//Process Speed Level
                		else if((parsing_state == PARSE_MOTOR_DIR_RCVD)){
                    			parsing_state = PARSE_SPD_LEVEL_RCVD;
                    			tempSpdLevel = serial_read_buf[0];
                		}
               			//Process Duration
                		else if((parsing_state == PARSE_SPD_LEVEL_RCVD)){
                    			parsing_state = PARSE_DURATION_RCVD;
                    			tempDuration = serial_read_buf[0];
                		}
                
                		//Process GPCA State ID
                		else if((parsing_state == PARSE_DURATION_RCVD)){
                    			parsing_state = PARSE_GPCA_STATE_RCVD;
                		}
                		//Process Reserved 5 (Motor State)
                		else if((parsing_state == PARSE_GPCA_STATE_RCVD)){
                    			parsing_state = PARSE_MOTOR_STATE_RCVD;
                    			tempMotorState = serial_read_buf[0];
                		}
                		//Process Reserved 6 (Remaining volume percent)
                		else if((parsing_state == PARSE_MOTOR_STATE_RCVD)){
                    			parsing_state = PARSE_RESERVE6_RCVD;
					tempRemainVol = serial_read_buf[0];
                    			//printf("Reserved 6 process");
                		}
                		//Process Reserved 7
                		else if((parsing_state == PARSE_RESERVE6_RCVD)){
                    			parsing_state = PARSE_RESERVE7_RCVD;
                    			//printf("Reserved 7 process");
                		}     
                		//Process Reserved 8
                		else if((parsing_state == PARSE_RESERVE7_RCVD)){
                    			parsing_state = PARSE_RESERVE8_RCVD;
                    			//printf("Reserved 8 process");
                		}   
                		//Process EOP
                		else if((parsing_state == PARSE_RESERVE8_RCVD) && serial_read_buf[0] == EOP ){
                    			parsing_state = PARSE_INIT;
					INFUSION_DURATION = tempDuration; 
					MOTOR_SPEED_LEVEL = tempSpdLevel;
					MOTOR_STATE = tempMotorState;
					MOTOR_DIRECTION = tempMotorDir;
					REMAIN_VOL_PERCENT = tempRemainVol;
					printf("Get a packet, Duration : %x \t Spd Level : %x \t Motor state : %x \t Motor Dir : %x Remain Vol : %d\t\n", INFUSION_DURATION, MOTOR_SPEED_LEVEL, MOTOR_STATE, MOTOR_DIRECTION, REMAIN_VOL_PERCENT); 
                		}else{
                    			parsing_state = PARSE_INIT;
					printf("Parsing error received : %x\n", serial_read_buf[0]);
                		}

			}//end if-else

		}// end while (read_err)

		close(serial_fd);
	}//end while

	close(serial_fd);		//Close serial port
	pthread_exit(NULL);

}//end thread function


/** \brief GPCA state machine thread : This thread shall be filled with source codes automatically generated from TIMES tool.
*	\param ptr not used.
*	\return none.
*/

void *gpca_state_machine_thread (void *ptr){
	printf("GPCA State Machine crated \n");

	///setClock
	setClock(GPCA_Instance_clk_x1,0);

	while(1){
		check_trans();	//This function will constantly check transition of GPCA automata
		printf("Current state : %d \t Current clock %ld \n", getCurrentState(), rdClock(GPCA_Instance_clk_x1));
		sleep(1);	//The transition will be checked at a predefined interval.
	}
	
}

/** \brief Clock Manager Thread : Provide clock source for GPCA-StateMachine.
*	\param ptr not used.
*	\return none.
*/
void *gpca_clock_manager (void *ptr){
	printf("Clock manager has been created...\n");
	
	while(1){
		sleep(1);
		GPCA_Instance_clk_x1 = (GPCA_Instance_clk_x1 + 1)% MAX_CLOCK_VALUE;
	}

	
}

/** \brief Observer adapter Thread : Provide GPCA-State to client at a regular interval.
*	\param ptr not used.
*	\return none.
*/

void *observer_adapter (void *ptr){
	int ret1;		
	printf("Observer adapter has been created...\n");
	while(1){
		if(clnt_socket < 0){
			printf("Client Socket is not estabilished...\n\r");
		}else{
			ret1 = send_gpca_pkt_to_client();
			if(ret1 >= 0){
				printf("Success in GPCA reporting...\n\r");
			}else{
				printf("Err in writing client socket...\n\r");
			}
	
		}
		sleep(1);
	}//end while
}

/**
 * Update the list of active transitions when a transition is taken,
 * and release task of target location (TIMES code).
 *
 * @param trn The transition taken
 */
static void clear_and_set ( unsigned char trn ) {
  int jj;
  // Clear outgoing transition from source location
  jj=trans[trn].from;
  do {
    trans[loc[jj]].active=0;
  } while(loc[++jj]<tid_offset);
  // Set outgoing transition from target location 
  jj=trans[trn].to;
  while ( loc[jj]<tid_offset ) {
    trans[loc[jj]].active=1;
    jj++;
  }
  // Release task of target location
  //releaseTask(loc[jj]-tid_offset);

  //TASK_RELEASED_HOOK(loc[jj]-tid_offset);
}


/**
 * Check if any of the active transitions are enabled, and if so take
 * it. Will continue until a stable state (no more enabled
 * transitions) is reached (TIMES code).
 *
 * @param data Unused (should be null).
 * @return false when in stable state
 */
void check_trans(){
	//printf("Called check_trans\n");
  int trn;
#if NB_CHAN>0
  int compl_trn;
#endif /* NB_CHAN>0 */
  for(trn=0; trn<NB_TRANS; trn++) {
    if( trans[trn].active ) {
      if( eval_guard(trn) ) {
#if NB_CHAN>0
	if( trans[trn].sync > -1 ) {
	  if( (compl_trn = 
	      check_synch( trans[trn].sync )) ) {
	    if( IS_SEND(trans[trn].sync) ) {
	      assign( trn );
	      assign( compl_trn );

	      //TRANS_TAKEN_HOOK(trans[trn].to);
	      //TRANS_TAKEN_HOOK(trans[compl_trn].to);

	    } else {
	      assign( compl_trn ); 
	      assign( trn );	    

	      //TRANS_TAKEN_HOOK(trans[compl_trn].to);
	      //TRANS_TAKEN_HOOK(trans[trn].to);

	    }
	    clear_and_set( trn );
	    clear_and_set( compl_trn );	   
	    trn=-1;
	  }
	} else 
#endif /* NB_CHAN>0 */ 
	{
	  assign( trn );
	  //TRANS_TAKEN_HOOK(trans[trn].to);
	  current_GPCA_state = trans[trn].to;		//Gluecode to keep track of the current GPCA state
	  printf("current trans : %d\t trans to : %d \n", trn, trans[trn].to);
	  
	  //If the state has been changed.
	  if(prev_GPCA_state != current_GPCA_state){
		//Do action on state change.
		printf("==>  GPCA state has been changed.\n");
		action_handler_on_state_change(prev_GPCA_state, current_GPCA_state);
		//Update the previous GPCA state.
		prev_GPCA_state = current_GPCA_state;
	  }
	  clear_and_set( trn );
	  trn=-1;
	} 
      }
    }
  }
  //return false;
}

/**
 * Evaluate guards on transition trn.
 * @param trn Transition id
 * @return true if guard satisfied, false otherwise.
 */
bool eval_guard(int trn) {
  switch(trn) {
  case 0: return (E_PauseInfusion==1);
  case 1: return (E_Cancel==1);
  case 2: return (E_ConfirmPauseInfusion==1);
  case 3: return (rdClock(GPCA_Instance_clk_x1)>=MAX_PAUSED_T);
  case 4: return (rdClock(GPCA_Instance_clk_x1)>=MAX_ALRM_T);
  case 5: return (E_ClearAlarm==1);
  case 6: return (E_StopInfusion==1);
  case 7: return (E_Cancel==1);
  case 8: return (E_ConfirmStopInfusion==1);
  case 9: return (Level_Two_Alarm==1);
  case 10: return (E_ClearAlarm==1);
  case 11: return (rdClock(GPCA_Instance_clk_x1)>=MAX_ALRM_T);
  case 12: return (MCond_6_6==1);
  case 14: return (Cond_6_3==1);
  case 15: return (E_CheckDrug==1);
  case 16: return (TCond_6_5==1);
  case 17: return (E_CheckDrug==1);
  case 18: return (rdClock(GPCA_Instance_clk_x1)>=MAX_CHECK_DRUG_T);
  case 19: return (TCond_6_4==1);
  case 20: return (E_ChangeDoseRate==1);
  case 21: return (E_Cancel==1);
  case 22: return (E_ConfirmDoseRate==1);
  case 23: return (Cond_7_4==1);
  case 24: return (Cond_7_3==1);
  case 26: return (E_ChangeVTBI==1);
  case 27: return (E_Cancel==1);
  case 28: return (rdClock(GPCA_Instance_clk_x1)>=MAX_WAIT_INPUT_T);
  case 30: return (Cond_4_1==1);
  case 31: return (E_ChangeVTBI==1);
  case 32: return (E_ConfirmVTBI==1);
  case 33: return (Cond_4_2==1);
  case 34: return (Cond_4_3==1);
  case 35: return (E_ConfirmVTBI==1);
  case 36: return (E_ConfirmVTBI==1);
  case 37: return (E_ChangeVTBI==1);
  case 38: return (E_ChangeDoseRate==1);
  case 39: return (E_Cancel==1);
  case 40: return (rdClock(GPCA_Instance_clk_x1)>=MAX_WAIT_INPUT_T);
  case 42: return (E_ConfirmDoseRate==1);
  case 43: return (Cond_4_4==1);
  case 44: return (E_ChangeDoseRate==1);
  case 45: return (Cond_4_5==1);
  case 46: return (E_ChangeDoseRate==1);
  case 47: return (Cond_4_6==1);
  case 48: return (E_ConfirmDoseRate==1);
  case 49: return (E_ConfirmDoseRate==1);
  case 50: return (rdClock(GPCA_Instance_clk_x1)>=MAX_WAIT_INPUT_T);
  case 52: return (E_StartInfusion==1);
  case 53: return (Cond_5_1==1);
  case 54: return (MCDR_In_Progress==0&&E_CheckAdminSet==1);
  case 55: return (Cond_3_1==1);
  case 56: return (E_CheckAdminSet_1==1);
  case 57: return (Cond_3_2==1);
  case 58: return (Cond_3_3==1);
  case 59: return (E_Prime==1);
  case 60: return (Cond_3_4==1);
  case 61: return (E_CheckDrug==1);
  case 62: return (Cond_3_5==1);
  case 63: return (E_CheckDrug==1);
  case 64: return (Cond_3_6==1);
  case 65: return (Cond_3_7==1);
  case 66: return (E_CheckDrug==1);
  case 67: return (Cond_3_8==1);
  case 68: return (Cond_3_9==1);
  case 69: return (E_CheckDrug==1);
  case 70: return (Cond_3_11==1);
  case 71: return (E_ConfirmConcentration==1);
  case 72: return (E_Cancel==1);
  case 73: return (E_CheckDrug==1);
  case 74: return (E_PowerButton==1);
  case 75: return (MCond_1_4==1);
  case 76: return (rdClock(GPCA_Instance_clk_x1)>=MAX_POST_WAIT);
  case 78: return (MCond_1_3==1);
  case 79: return (Cond_3_10==1);
  case 80: return (E_ConfigureInfusionProgram==1);
  case 81: return (Cond_5_2==1);
  case 82: return (E_NewInfusion==1);
  case 83: return (E_Cancel==1);
  case 84: return (Cond_6_2==1);
  case 85: return (E_StopInfusion==1);
  case 86: return (E_Cancel==1);
  case 87: return (E_ConfirmStopInfusion==1);
  case 88: return (E_StopInfusion==1);
  case 89: return (Level_Two_Alarm==1);
  case 90: return (E_StopInfusion==1);
  case 91: return (Level_Two_Alarm==1);
  case 92: return (Level_Two_Alarm==1);
  case 93: return (E_ChangeDoseRate==1);
  case 94: return (Cond_6_2==1);
  case 95: return (Cond_6_2==1);
  case 96: return (E_Cancel==1);
  case 97: return (Level_Two_Alarm==1);
  case 98: return (Cond_6_2==1);
  case 99: return (rdClock(GPCA_Instance_clk_x1)>=MAX_WAIT_INPUT_T);
  case 101: return (Cond_6_2==1);
  case 102: return (E_StopInfusion==1);
  case 103: return (E_StopInfusion==1);
  case 104: return (E_RequestBolus==1);
  case 105: return (Cond_7_1==1);
  case 106: return (Cond_7_2==1);
  case 13:
  case 25:
  case 29:
  case 41:
  case 51:
  case 77:
  case 100:
    return true;
  }
  return false;
}

/**
 * Perform assignments on transition trn.
 * @param trn Transition id.
 */

void assign(int trn) {
  switch(trn) {
  case 0: 
    E_PauseInfusion=0; break;
  case 1: 
    E_Cancel=0; break;
  case 2: 
    E_ConfirmPauseInfusion=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 3: 
    setClock(GPCA_Instance_clk_x1,0); break;
  case 5: 
    E_ClearAlarm=0; break;
  case 6: 
    E_StopInfusion=0; break;
  case 7: 
    E_Cancel=0; break;
  case 8: 
    E_ConfirmStopInfusion=0; break;
  case 9: 
    setClock(GPCA_Instance_clk_x1,0);
    Level_Two_Alarm=0; break;
  case 10: 
    E_ClearAlarm=0; break;
  case 14: 
    setClock(GPCA_Instance_clk_x1,0);
    Cond_6_3=0; break;
  case 15: 
    E_CheckDrug=0; break;
  case 17: 
    E_CheckDrug=0; break;
  case 20: 
    E_ChangeDoseRate=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 21: 
    E_Cancel=0; break;
  case 22: 
    E_ConfirmDoseRate=0; break;
  case 23: 
    Cond_7_4=0; break;
  case 24: 
    Cond_7_3=0; break;
  case 26: 
    E_ChangeVTBI=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 27: 
    E_Cancel=0; break;
  case 28: 
    setClock(GPCA_Instance_clk_x1,0); break;
  case 30: 
    Cond_4_1=0; break;
  case 31: 
    E_ChangeVTBI=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 32: 
    E_ConfirmVTBI=0; break;
  case 33: 
    Cond_4_2=0; break;
  case 34: 
    Cond_4_3=0; break;
  case 35: 
    E_ConfirmVTBI=0; break;
  case 36: 
    E_ConfirmVTBI=0; break;
  case 37: 
    E_ChangeVTBI=0; break;
  case 38: 
    setClock(GPCA_Instance_clk_x1,0);
    E_ChangeDoseRate=0; break;
  case 39: 
    E_Cancel=0; break;
  case 40: 
    setClock(GPCA_Instance_clk_x1,0); break;
  case 42: 
    E_ConfirmDoseRate=0; break;
  case 43: 
    Cond_4_4=0; break;
  case 44: 
    E_ChangeDoseRate=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 45: 
    Cond_4_5=0; break;
  case 46: 
    E_ChangeDoseRate=0; break;
  case 47: 
    setClock(GPCA_Instance_clk_x1,0);
    Cond_4_6=0; break;
  case 48: 
    E_ConfirmDoseRate=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 49: 
    E_ConfirmDoseRate=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 50: 
    setClock(GPCA_Instance_clk_x1,0); break;
  case 52: 
    E_StartInfusion=0; break;
  case 53: 
    Cond_5_1=0; break;
  case 54: 
    MCDR_In_Progress=1;
    E_CheckAdminSet=0; break;
  case 55: 
    Cond_3_1=0; break;
  case 56: 
    E_CheckAdminSet_1=0; break;
  case 57: 
    Cond_3_2=0; break;
  case 58: 
    Cond_3_3=0; break;
  case 59: 
    E_Prime=0; break;
  case 60: 
    Cond_3_4=0; break;
  case 61: 
    E_CheckDrug=0; break;
  case 62: 
    Cond_3_5=0; break;
  case 63: 
    E_CheckDrug=0; break;
  case 64: 
    Cond_3_6=0; break;
  case 65: 
    Cond_3_7=0; break;
  case 66: 
    E_CheckDrug=0; break;
  case 67: 
    Cond_3_8=0; break;
  case 68: 
    Cond_3_9=0; break;
  case 69: 
    E_CheckDrug=0; break;
  case 70: 
    Cond_3_11=0; break;
  case 71: 
    E_ConfirmConcentration=0; break;
  case 72: 
    E_Cancel=0; break;
  case 73: 
    E_CheckDrug=0; break;
  case 74: 
    E_PowerButton=0;
    Power_ON=1;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 75: 
    MCond_1_4=0; break;
  case 78: 
    MCond_1_3=0; break;
  case 79: 
    Cond_3_10=0; break;
  case 80: 
    MCDR_In_Progress=0;
    E_ConfigureInfusionProgram=0; break;
  case 81: 
    Cond_5_2=0; break;
  case 82: 
    E_NewInfusion=0; break;
  case 83: 
    E_Cancel=0; break;
  case 84: 
    Cond_6_2=0; break;
  case 85: 
    E_StopInfusion=0; break;
  case 86: 
    E_Cancel=0; break;
  case 87: 
    E_ConfirmStopInfusion=0; break;
  case 88: 
    E_StopInfusion=0; break;
  case 89: 
    Level_Two_Alarm=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 90: 
    E_StopInfusion=0; break;
  case 91: 
    Level_Two_Alarm=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 92: 
    Level_Two_Alarm=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 93: 
    E_ChangeDoseRate=0; break;
  case 94: 
    Cond_6_2=0; break;
  case 95: 
    Cond_6_2=0; break;
  case 96: 
    E_Cancel=0; break;
  case 97: 
    Level_Two_Alarm=0;
    setClock(GPCA_Instance_clk_x1,0); break;
  case 98: 
    Cond_6_2=0; break;
  case 99: 
    setClock(GPCA_Instance_clk_x1,0); break;
  case 101: 
    Cond_6_2=0; break;
  case 102: 
    E_StopInfusion=0; break;
  case 103: 
    E_StopInfusion=0; break;
  case 104: 
    E_RequestBolus=0; break;
  case 105: 
    Cond_7_1=0; break;
  case 106: 
    Cond_7_2=0; break;

  }
}

/**
*	\brief implementation of [setClock] interface in TIMES code.
*	\param clockID clock id used in the automata.
*	\param value clock value to be newly set.
*/
void setClock(time_t clockID, long value){
	GPCA_Instance_clk_x1 = value;
}

/**
*	\brief implementation of [rdClock] interface in TIMES code.
*	\param clockID clock id used in the automata.
*	\return the current clock value of clockID.
*/
time_t rdClock(time_t clockID){

	return GPCA_Instance_clk_x1;	//Return current clock value
}

/**
* \brief Get the current GPCA state.
* \return the current GPCA state.
*/
int getCurrentState(void){
	return current_GPCA_state;
}

/**
*  \brief Send GPCA packet to client through TCP/IP.
*  \return 1 on success, -1 on fail.
*/
int send_gpca_pkt_to_client(void)
{
	int i, write_res;
        pkt_gpca[INDEX_SOP] = SOP;
        pkt_gpca[INDEX_MOTOR_DIR] = MOTOR_DIRECTION;
        pkt_gpca[INDEX_SPD_LEVEL] = MOTOR_SPEED_LEVEL;
        pkt_gpca[INDEX_DURATION] = INFUSION_DURATION;
	pkt_gpca[INDEX_GPCA_STATE] = getCurrentState();
        pkt_gpca[INDEX_MOTOR_STATE] = MOTOR_STATE;
	pkt_gpca[INDEX_CHANNEL_ACK] = CHAN_NO_ACK; 
        pkt_gpca[INDEX_EOP] = EOP;

	pkt_gpca[INDEX_REMAIN_VOL_PERCENT] = REMAIN_VOL_PERCENT;
	
        //Send packet to ethernet
        for(i = 0 ; i < PKT_SIZE ; i++){
		if(clnt_socket < 0){
			printf("Client Socket is not estabilished...\n\r");
		}else{
			write_res = write(clnt_socket, &pkt_gpca[i], 1);
			if(write_res >= 0){
				printf("%X ", pkt_gpca[i]);
			}else{
				printf("Err in writing client socket...\n\r");
				return -1;
			}
		}
	}// end for
	return 1;
}

/**
* @brief Send acknowledgement for processing channels.
*/

int send_channel_ack_to_client(void){
	int i, write_res;
        pkt_gpca[INDEX_SOP] = SOP;
        pkt_gpca[INDEX_MOTOR_DIR] = MOTOR_DIRECTION;
        pkt_gpca[INDEX_SPD_LEVEL] = MOTOR_SPEED_LEVEL;
        pkt_gpca[INDEX_DURATION] = INFUSION_DURATION;
	pkt_gpca[INDEX_GPCA_STATE] = getCurrentState();
        pkt_gpca[INDEX_MOTOR_STATE] = MOTOR_STATE;
	pkt_gpca[INDEX_CHANNEL_ACK] = CHAN_ACK; 
        pkt_gpca[INDEX_EOP] = EOP;
	
        //Send packet to ethernet
        for(i = 0 ; i < PKT_SIZE ; i++){
		if(clnt_socket < 0){
			printf("Client Socket is not estabilished...\n\r");
		}else{
			write_res = write(clnt_socket, &pkt_gpca[i], 1);
			if(write_res >= 0){
				printf("%X ", pkt_gpca[i]);
			}else{
				printf("Err in writing client socket...\n\r");
				return -1;
			}
		}
	}// end for
	printf("=========Channel ack sent==========\n");
	return 1;
}


/**
* @brief Activate an gpca channel that interacts with environment.
* @Param cmd Channel Indicator received from Environment.
* @Return The channel id that has been activated, -1 in case of invalid channel.
*/

int activate_gpca_channel(unsigned char cmd){

	int ret_val = 0;
	//#define CMD_EVT_RequestBolus 0x40
	if(cmd == CMD_EVT_RequestBolus){
		//Not implemented at the model level yet.
	
	}
	//#define CMD_EVT_StopInfusion 0x41
	else if (cmd == CMD_EVT_StopInfusion){
		if(getCurrentState() == GPCA_Instance_Infusion_NormalOperation || getCurrentState() == GPCA_Instance_InfusionPaused || getCurrentState() == GPCA_Instance_Alrm_TooLongInfusionPause || getCurrentState() == GPCA_Instance_Alrm_LevelTwoHardwareFailure || getCurrentState() == GPCA_Instance_Alrm_WrongDrug || getCurrentState() == GPCA_Instance_Alrm_EmptyReservior){
			E_StopInfusion = 1;
			return cmd;
		}
	}
	//#define CMD_EVT_PowerButton 0x42
	else if (cmd == CMD_EVT_PowerButton){
		if(getCurrentState() == GPCA_Instance_POST_Init){
			printf("PowerButton Received\n");
			 E_PowerButton = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_PauseInfusion 0x43	
	else if (cmd == CMD_EVT_PauseInfusion){
		if(getCurrentState() == GPCA_Instance_Infusion_NormalOperation){
			 E_PauseInfusion = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_ConfirmPauseInfusion 0x44
	else if (cmd == CMD_EVT_ConfirmPauseInfusion){
		if(getCurrentState() == GPCA_Instance_ConfirmPause){
			 E_ConfirmPauseInfusion = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_ClearAlarm 0x45
	else if (cmd == CMD_EVT_ClearAlarm){
		if(getCurrentState() == GPCA_Instance_Alrm_TooLongInfusionPause || getCurrentState() == GPCA_Instance_Alrm_LevelTwoHardwareFailure){
			 E_ClearAlarm = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_Cancel 0x46
	else if (cmd == CMD_EVT_Cancel){ //ConfirmPause, ConfirmStop, ChangeRate, ChangeVTBI, ChangeDoseRate, DisplayPatientProfile
		if(getCurrentState() == GPCA_Instance_ConfirmPause || getCurrentState() == GPCA_Instance_InfusionPaused || getCurrentState() == GPCA_Instance_PausedStopConfirm || getCurrentState() == GPCA_Instance_ConfirmStop || getCurrentState() == GPCA_Instance_ChangeRate || getCurrentState() == GPCA_Instance_Alrm_UnsafeNewRate || getCurrentState() == GPCA_Instance_ChangeVTBI || getCurrentState() == GPCA_Instance_ChangeDoseRate || getCurrentState() == GPCA_Instance_DisplayPatientProfile ){
			 E_Cancel = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_ChangeDoseRate 0x47
	else if (cmd == CMD_EVT_ChangeDoseRate){ //DisplayDoseRate, Alarming_DoseRate_OutOfBound,Wrn_DoseRateSoftLimitsViolation,  Infusion_NormalOperation, 
		if(getCurrentState() == GPCA_Instance_DisplayDoseRate || getCurrentState() == GPCA_Instance_Infusion_NormalOperation || getCurrentState() == GPCA_Instance_Alarming_DoseRate_OutOfBound || getCurrentState() == GPCA_Instance_Wrn_DoseRateSoftLimitsViolation){
			 E_ChangeDoseRate = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_ConfigureInfusionProgram 0x48
	else if (cmd == CMD_EVT_ConfigureInfusionProgram){ //DisplayPatientProfile
		if(getCurrentState() == GPCA_Instance_DisplayPatientProfile){
			 E_ConfigureInfusionProgram = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_ConfirmStopInfusion 0x49
	else if (cmd == CMD_EVT_ConfirmStopInfusion){ //ConfirmStop, PausedStopConfirm
		if(getCurrentState() == GPCA_Instance_ConfirmStop || getCurrentState() == GPCA_Instance_PausedStopConfirm){
			 E_ConfirmStopInfusion = 1;
			return cmd;
		}
	}
	//#define CMD_EVT_ConfirmDoseRate 0x4A
	else if (cmd == CMD_EVT_ConfirmDoseRate){ //ChangeRate, DisplayDoseRate, ChangeDoseRate, Wrn_DoseRateSoftLimitsViolation, 
		if(getCurrentState() == GPCA_Instance_ChangeRate || getCurrentState() == GPCA_Instance_DisplayDoseRate || getCurrentState() == GPCA_Instance_ChangeDoseRate || getCurrentState() == GPCA_Instance_Wrn_DoseRateSoftLimitsViolation){
			 E_ConfirmDoseRate = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_NewInfusion 0x4B
	else if (cmd == CMD_EVT_NewInfusion){ //InfusionStopped
		if(getCurrentState() == GPCA_Instance_InfusionStopped){
			 E_NewInfusion = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_ChangeVTBI 0x4C
	else if (cmd == CMD_EVT_ChangeVTBI){ //DisplayVTBI, Alrm_VTBIHardLimitsViolation, Wrn_VTBISoftLimitsViolation,
		if(getCurrentState() == GPCA_Instance_DisplayVTBI || getCurrentState() == GPCA_Instance_Alrm_VTBIHardLimitsViolation || getCurrentState() == GPCA_Instance_Wrn_VTBISoftLimitsViolation){
			 E_ChangeVTBI = 1;
			return cmd;
		}
	}
	//#define CMD_EVT_ConfirmVTBI 0x4D
	else if (cmd == CMD_EVT_ConfirmVTBI){ //ChangeVTBI, DisplayVTBI, Wrn_VTBISoftLimitsViolation
		if(getCurrentState() == GPCA_Instance_ChangeVTBI || getCurrentState() == GPCA_Instance_DisplayVTBI || getCurrentState() == GPCA_Instance_Wrn_VTBISoftLimitsViolation){
			 E_ConfirmVTBI = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_StartInfusion 0x4E
	else if (cmd == CMD_EVT_StartInfusion){ //DisplaySetting
		if(getCurrentState() == GPCA_Instance_DisplaySetting){
			 E_StartInfusion = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_CheckAdminSet 0x4F
	else if (cmd == CMD_EVT_CheckAdminSet){  //CDR_Init
		if(getCurrentState() == GPCA_Instance_CDR_Init){
			 MCDR_In_Progress = 0; //TODO Modified.
			 E_CheckAdminSet = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_CheckAdminSet_1 0x50
	else if (cmd == CMD_EVT_CheckAdminSet_1){ //Alrm_IncorrectAdministrationSetup
		if(getCurrentState() == GPCA_Instance_Alrm_IncorrectAdministrationSetup){
			 E_CheckAdminSet_1 = 1;
			return cmd;
		}	
	}
	//#define CMD_EVT_CheckDrug 0x51
	else if (cmd == CMD_EVT_CheckDrug){ //DisplayDrugInfo, Alrm_UnknownDrug, Alrm_DoseUnitMismatch , Wrn_ConcentrationSoftLimitsViolation, Alrm_ConcentrationHardLimitsViolation, 
		if(getCurrentState() == GPCA_Instance_DisplayDrugInfo || getCurrentState() == GPCA_Instance_Alrm_UnknownDrug || getCurrentState() == GPCA_Instance_Alrm_DoseUnitMismatch || getCurrentState() == GPCA_Instance_Wrn_ConcentrationSoftLimitsViolation || getCurrentState() == GPCA_Instance_Alrm_ConcentrationHardLimitsViolation){
			 E_CheckDrug = 1;
			return cmd;
		}	 
	}
	//#define CMD_EVT_ConfirmConcentration 0x52
	else if (cmd == CMD_EVT_ConfirmConcentration){ //Wrn_ConcentrationSoftLimitsViolation
		if(getCurrentState() == GPCA_Instance_Wrn_ConcentrationSoftLimitsViolation){
			E_ConfirmConcentration = 1;
			return cmd;
		}
	}
	//#define CMD_EVT_Prime 0x53
	else if (cmd == CMD_EVT_Prime){ //Alrm_PrimeFailure   
		if(getCurrentState() == GPCA_Instance_Alrm_PrimeFailure){
			 E_Prime = 1;
			return cmd;
		}
	}
	//#define CMD_V_MCond_1_4=0;  //added.
	else if(cmd == CMD_V_MCond_1_4){
		if(getCurrentState() == GPCA_Instance_Post_In_Progress){
			MCond_1_4 = 1;
			MCDR_In_Progress = 0;	//TODO this needs to be modified.
			return cmd;
		}
	}
	//#define CMD_V_Cond_3_2 0x86
	else if(cmd == CMD_V_Cond_3_2){
		if(getCurrentState() == GPCA_Instance_CheckAdministrationSet){
			Cond_3_2 = 1;		
			return cmd;
		}
	}
	//#define CMD_V_Cond_3_4 0x88
	else if(cmd == CMD_V_Cond_3_4){
		if(getCurrentState() == GPCA_Instance_CheckPrime){
			Cond_3_4 = 1;
			return cmd;
		}
	}
	//#define CMD_V_Cond_3_6 0x8A
	else if(cmd == CMD_V_Cond_3_6){
		if(getCurrentState() == GPCA_Instance_CheckDrugType){
			Cond_3_6 = 1;
			return cmd;
		}
	}
	//#define CMD_V_Cond_3_8 0x8C
	else if(cmd == CMD_V_Cond_3_8){
		if(getCurrentState() == GPCA_Instance_CheckDoseUnit){
			Cond_3_8 = 1;
			return cmd;
		}
	}
	//#define CMD_V_Cond_3_11 0x8E
	else if(cmd == CMD_V_Cond_3_11){
		if(getCurrentState() == GPCA_Instance_CheckConcentration){
			Cond_3_11 = 1;
			return cmd;		
		}
	}
	//#define CMD_V_Cond_4_3 0x7F
	else if(cmd == CMD_V_Cond_4_3){
		if(getCurrentState() == GPCA_Instance_CheckVTBI){
			Cond_4_3 = 1;
			return cmd;		
		}	
	}
	//#define CMD_V_Cond_4_6 0x82
	else if(cmd == CMD_V_Cond_4_6){
		if(getCurrentState() == GPCA_Instance_CheckDoseRate){
			Cond_4_6 = 1;
			return cmd;
		}
	}
	//#define CMD_V_Cond_5_2=0;   //added.
	else if(cmd == CMD_V_Cond_5_2){
		if(getCurrentState() == GPCA_Instance_ReadyToStart){
			Cond_5_2 = 1;
			return cmd;		
		}
	}

	return -1;
}
/**
* @brief An action handler when state change is reported by check_trans fucntion, it is required to do proper action depending on the state that goes into.
* @Param prev_state a previous gpca state.
* @Param current_state a current gpca state.
* @Return The result of processing the corresponding action handler, -1 in case of action handler fails to do a proper action.
*/
int action_handler_on_state_change(int prev_state, int current_state){
	
	if(prev_state == GPCA_Instance_POST_Init){
		if(current_state == GPCA_Instance_Post_In_Progress){
			printf(">>Do Post In Progress\n");
			return 1;
		}
	}
	//if(prev_state == GPCA_Instance_ISSM_Init){
	if(prev_state == GPCA_Instance_ReadyToStart){
		if(current_state == GPCA_Instance_Infusion_NormalOperation){
			printf(">>Start Motor\n");
			write_byte_to_controller(serial_fd, CONT_CMD_TOGGLE_MOTOR_OPERATION); 
			return 1;
		}
	}
	if(prev_state == GPCA_Instance_ConfirmPause){
		if(current_state == GPCA_Instance_InfusionPaused){
			printf(">>Pause Motor\n");
			write_byte_to_controller(serial_fd, CONT_CMD_TOGGLE_MOTOR_OPERATION); 
			return 1;
		}
	}
	if(prev_state == GPCA_Instance_Alrm_TooLongInfusionPause){
		if(current_state == GPCA_Instance_Infusion_NormalOperation){
			printf(">>Start Motor\n");
			//Start motor
			write_byte_to_controller(serial_fd, CONT_CMD_TOGGLE_MOTOR_OPERATION); 
			//Clear Alarm
			write_byte_to_controller(serial_fd, CONT_CMD_ALARM_BUZZ_OFF);
			write_byte_to_controller(serial_fd, CONT_CMD_ALARM_LED_OFF);
			return 1;
		}
	}
	if(prev_state == GPCA_Instance_InfusionPaused){
		if(current_state == GPCA_Instance_Alrm_TooLongInfusionPause){
			printf(">>Alarm !! Too Long Infused\n");
			//Raise alarm.
			write_byte_to_controller(serial_fd, CONT_CMD_ALARM_BUZZ_ON);
			write_byte_to_controller(serial_fd, CONT_CMD_ALARM_LED_ON);
			return 1;
		}
	}
	if(prev_state == GPCA_Instance_Alrm_TooLongInfusionPause){
		if(current_state == GPCA_Instance_InfusionStopped){
			printf(">>Infusion Session is stopped\n");
			//Clear Alarm
			write_byte_to_controller(serial_fd, CONT_CMD_ALARM_BUZZ_OFF);
			write_byte_to_controller(serial_fd, CONT_CMD_ALARM_LED_OFF);
		}
	}
	return 0;
}

/**
* @brief Send one-byte data to the infusion controller through RS232.
* @Param fd file descriptor of the RS232 connection to the infusion controller.
* @Param data command to send to the infusion controller.
* @Return The result of sending command to the infusion controller, -1 in case fails to send the commmand.
*/
int write_byte_to_controller(int fd, char data){
	int write_res = 0;
	write_res = write(fd, &data, 1);
	if(write_res <= 0){
		printf("Error in writing to infusion controller...\n\r");	
		return 0;
	}else{
		printf("[%c] has been delivered to infusion controller\n\r", data);
		return 1;
	}
}
